/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/08/22.
//

#include "grid.hpp"

#include "fggl/assets/loader.hpp"
#include "fggl/entity/gridworld/zone.hpp"

#include "fggl/gui/model/parser.hpp"
#include "fggl/gui/renderer/renderer.hpp"

using namespace fggl::gfx::colours;

namespace demo {

	using namespace fggl::entity::grid;

	static void build_tileset(TileSet& tiles) {
		tiles.m_floors.push_back(FloorTile{FloorTile::IMPOSSIBLE, BLACK});
		tiles.m_floors.push_back(FloorTile{1, GREEN});
		tiles.m_floors.push_back(FloorTile{1, YELLOW_GREEN});

		fggl::entity::grid::WallTile noWall{};
		tiles.m_walls.push_back(noWall);

		fggl::entity::grid::WallTile solidWall{
			.render = true,
			.colour = DARK_SLATE_GRAY
		};
		tiles.m_walls.push_back(solidWall);
	}

	static void build_room(DemoGrid* area, fggl::math::vec2i center, fggl::math::vec2i size) {

		for (int yOffset = -size.y; yOffset <= size.y; ++yOffset) {
			auto yPos = yOffset + center.y;
			area->setWallAt(center.x - size.x, yPos, false, 1);
			area->setWallAt(center.x + size.x + 1, yPos, false, 1);
		}

		for (int xOffset = -size.x; xOffset <= size.x; ++xOffset) {
			auto xPos = center.x + xOffset;

			area->setWallAt(xPos, center.y - size.y, true, 1);
			area->setWallAt(xPos, center.y + size.y + 1, true, 1);

			for (int yOffset = -size.y; yOffset <= size.y; ++yOffset) {
				auto yPos = yOffset + center.y;
				area->setFloorAt(xPos, yPos, 1);
			}
		}
	}

	void build_doorway(DemoGrid* area, fggl::math::vec2i position, bool north, int size = 1) {
		for ( auto offset = 0; offset < size; offset++) {
			if ( north ) {
				area->setWallAt(position.x + offset, position.y, north, 0);
			} else {
				area->setWallAt(position.x, position.y + offset, north, 0);
			}
		}
	}

	static fggl::entity::EntityID build_test_env(DemoGrid* area, LevelRules& rules) {
		area->clear();
		build_room(area, {5, 5}, {4,4});
		build_room(area, {11, 5}, {1,1});
		build_room(area, {17, 5}, {4,4});
		build_doorway(area, {10, 5}, false, 1);
		build_doorway(area, {13, 5}, false, 1);

		// set goal
		area->setFloorAt(17, 5, 2);

		// level rules
		rules.startingPower = 20;
		rules.startingPos = {5, 5};
		rules.startingDirection = 1;

		// player
		fggl::entity::EntityID player = fggl::entity::INVALID;
		auto& manager = area->entities();
		{
			player = manager.create();
			auto& cellPos = manager.add<CellPos>(player);
			auto& robotState = manager.add<RobotState>(player);
		}

		return player;
	}

	struct Action {
		const char* name;
		std::function<void(void)> callback;
	};

	GridScene::GridScene(fggl::App &app) : GameBase(app), m_tiles(), m_animator(15.0F), m_grid(nullptr), m_canvas() {
		m_animator.add([this](){this->tickPlayer();});

		auto btnGrid = std::make_unique<fggl::gui::GridBox>(0, 2);

		std::array<Action, 4> actions{{
										  {"<", [this]() { this->rotate(true); }},
										  {">", [this]() { this->rotate(false); }},
										  {"^", [this]() { this->forward(); }},
										  {"Z", [this]() {  } }
									  }};

		fggl::math::vec2i pos{0, 0};
		for (auto& action : actions) {
			fggl::math::vec2i size{32, 32};
			auto btn = std::make_unique<fggl::gui::Button>(pos, size);
			btn->label(action.name);
			btn->addCallback([=](){
				this->m_program.m_instructions.push_back({action.name, action.callback});
			});
			btnGrid->add(std::move(btn));
		}

		// control buttons
		{
			fggl::math::vec2i size{64, 32};
			auto btn = std::make_unique<fggl::gui::Button>(pos, size);
			btn->label("go");
			btn->addCallback([this](){
				if ( !this->m_program.playing ) {
					resetPuzzle();
					this->m_program.m_currInstruction = 0;
					this->m_program.playing = true;
				}
			});
			btnGrid->add(std::move(btn));
		}

		{
			fggl::math::vec2i size{64, 64};
			auto btn = std::make_unique<fggl::gui::Button>(pos, size);
			btn->label("Del");
			btn->addCallback([this](){
				if ( !this->m_program.playing ) {
					if ( !m_program.m_instructions.empty() ) {
						m_program.m_instructions.pop_back();
					}
				}
			});
			btnGrid->add(std::move(btn));
		}

		btnGrid->layout();
		m_canvas.add(std::move(btnGrid));

		// create a timeline panel
		std::unique_ptr<robot::Timeline> timeline = std::make_unique<robot::Timeline>(m_program);
		timeline->size({50,700}, {250, 250});
		m_canvas.add(std::move(timeline));

		m_canvas.layout();
	}

	void GridScene::activate() {
		GameBase::activate();
		fggl::debug::log(fggl::debug::Level::info, "GridScene::activate()");

		m_animator.reset();

		// fake loading the tileset
		if ( m_tiles.m_floors.empty() ) {
			build_tileset(m_tiles);
		}

		// create the grid world
		m_grid = std::make_unique<DemoGrid>(m_tiles);
		m_player = build_test_env(m_grid.get(), m_levelRules);

		resetPuzzle();
	}

	void GridScene::deactivate() {
		m_grid = nullptr;
	}

	constexpr float DRAW_SIZE = 64.0F;
	constexpr float DRAW_HALF = DRAW_SIZE / 2.0F;
	constexpr float WALL_HALF = 2.5F;

	static void render_grid(fggl::gfx::Paint& paint, fggl::entity::grid::Area2D<255, 255>& grid, fggl::App& app) {
		const fggl::math::vec2 wallOffsetNorth {0, -DRAW_HALF};
		const fggl::math::vec2 wallOffsetWest {-DRAW_HALF, 0};

		for (int i=0; i <= 31; ++i) {
			for (int j=0; j <= 31; ++j) {
				const fggl::math::vec2 drawPos{i * DRAW_SIZE, j*DRAW_SIZE};

				// draw floor
				{
					auto &floor = grid.floorAt(i, j);
					fggl::gfx::Path2D tileGfx = fggl::gfx::make_rect(drawPos, {DRAW_HALF, DRAW_HALF}, floor.colour);
					paint.fill(tileGfx);
				}

				auto& wallNorth = grid.wallAt(i, j, true);
				if ( wallNorth.render ) {
					fggl::gfx::Path2D tileGfx = fggl::gfx::make_rect(drawPos + wallOffsetNorth, {DRAW_HALF, WALL_HALF}, wallNorth.colour);
					paint.fill(tileGfx);
				}

				auto& wallWest = grid.wallAt(i, j, false);
				if ( wallWest.render ) {
					fggl::gfx::Path2D tileGfx = fggl::gfx::make_rect(drawPos + wallOffsetWest, {WALL_HALF, DRAW_HALF}, wallWest.colour);
					paint.fill(tileGfx);
				}
			}
		}

		// UI test
		auto widgetFactory = app.service<fggl::gui::model::WidgetFactory>();

		auto widget = widgetFactory->buildEmpty();
		widget->set("position", fggl::math::vec2{200.0F, 100.F});
		widget->set("size", fggl::math::vec2{500.0F, 300.F});
		widget->set("colour", fggl::gfx::colours::BLANCHED_ALMOND);
		fggl::gui::model::attr_box_set(*widget, "padding", 5.0F);

		auto handle = widgetFactory->buildEmpty();
		handle->set("border::bottom",5.0F);
		handle->set("position", fggl::math::vec2{0.0F, 0.0F});
		//handle->set("size", fggl::math::vec2{INFINITY, 50.0F});
		handle->set("text", "hello, world!");
		handle->set("colour", fggl::gfx::colours::ORANGE);
		fggl::gui::model::attr_box_set(*handle, "padding", 5.0F);
		widget->addChild(*handle);
		delete handle;

		auto content = widgetFactory->buildEmpty();
		content->set("position", fggl::math::vec2{0.0F, 50.0F});
		//content->set("size", fggl::gui::model::UNDEFINED_SIZE);
		content->set("colour", fggl::gfx::colours::BURLYWOOD);
		widget->addChild(*content);
		delete content;

		fggl::gui::renderer::layout(*widget);
		fggl::gui::renderer::visit(*widget, paint);
	}


	static void render_objects(fggl::gfx::Paint& paint, DemoGrid& grid) {
		auto& manager = grid.entities();

		auto entities = manager.find<CellPos>();
		for (const auto& entity : entities ) {
			//auto& sprite = manager.get<Sprite>(entity);
			auto& cellPos = manager.get<CellPos>(entity);

			// convert grid pos to world pos
			fggl::math::vec2f drawPos = cellPos.pos;
			drawPos *= DRAW_SIZE;
			drawPos += cellPos.drawOffset;

			fggl::gfx::ShapeOpts opts;
			opts.angleOffset = (float)cellPos.direction * fggl::math::HALF_PI + cellPos.rotationOffset;

			auto shape = fggl::gfx::make_shape(drawPos, DRAW_HALF, 3, NAVAJO_WHITE, opts);
			paint.fill(shape);
		}
	}

	static void update_canvas(fggl::input::Input& inputs, fggl::gui::Container& canvas) {
		fggl::math::vec2f cursorPos {
			inputs.mouse.axis(fggl::input::MouseAxis::X),
			inputs.mouse.axis(fggl::input::MouseAxis::Y)
		};

		// in canvas space
		fggl::math::vec2 projected {
			fggl::math::rescale_ndc(cursorPos.x, 0, 1920.f),
			fggl::math::rescale_ndc(cursorPos.y, 0, 1080.0f)
		};
		canvas.onMouseOver(projected);

		// detect clicks
		if (inputs.mouse.pressed(fggl::input::MouseButton::LEFT)) {
			auto* widget = canvas.getChildAt(projected);
			if (widget != nullptr) {
				fggl::debug::info("Button clicked");
				widget->activate();
			}
		}
	}

	void GridScene::update(float deltaTime) {
		GameBase::update(deltaTime);
		m_animator.update(deltaTime);

		m_canvas.update(deltaTime);
		update_canvas(input(), m_canvas);
	}

	void GridScene::tickPlayer() {
		if ( !m_program.playing ){
			return;
		}

		if ( m_program.m_currInstruction < m_program.m_instructions.size() ) {
			auto& robotState = m_grid->entities().get<RobotState>(m_player);
			if (robotState.power == 0) {
				// a looser is you
				resetPuzzle();
				return;
			}

			robotState.power--;
			m_program.step();
		} else {
			m_program.stop();
			checkVictory();
		}
	}

	void GridScene::resetPuzzle() {
		// reset instruction panel
		m_program.stop();

		// reset robot state
		auto& robotPos = m_grid->entities().get<CellPos>(m_player);
		auto& robotState = m_grid->entities().get<RobotState>(m_player);

		robotPos.pos = m_levelRules.startingPos;
		robotPos.direction = m_levelRules.startingDirection;
		robotState.power = m_levelRules.startingPower;
	}

	void GridScene::checkVictory() {
		if ( !m_program.playing ) {
			auto& botPos = m_grid->entities().get<CellPos>(m_player).pos;
			auto gridCell = m_grid->floorAt(botPos.x, botPos.y);
			if ( gridCell.colour == YELLOW_GREEN ) {
				// a winner is you!
				returnToMenu();
			}
		}
	}

	void GridScene::render(fggl::gfx::Graphics &gfx) {
		fggl::gfx::Paint paint;
		render_grid(paint, *m_grid, owner());
		render_objects(paint, *m_grid);

		m_canvas.render(paint);
		gfx.draw2D(paint);
	}

} // namespace demo
